# ======================================================================================
#
#     Rapid Deep Neural Networks
#
#     Licensed under the MIT License
# ______________________________________________________________________________________
# ......................................................................................

# Copyright (c) 2018-2026 Pantelis I. Kaplanoglou

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# .......................................................................................
import importlib
from PIL import Image
from .fileobject import FileObject


from PIL import Image
import numpy as np

class PNGFile(FileObject):
  # ----------------------------------------------------------------------------------
  def __init__(self, filename, parent_folder=None, error_template=None):
    super(PNGFile, self).__init__(filename, parent_folder, error_template, "png")
  # ----------------------------------------------------------------------------------
  def load(self, filename=None):
    filename = self._useFileName(filename)
    img = Image.open(filename).convert("RGB")
    return np.array(img).astype("uint8")
  # ----------------------------------------------------------------------------------
  def save(self, image_ndarray, filename=None):
    filename = self._useFileName(filename)

    if image_ndarray.dtype != np.uint8:
      image_ndarray = image_ndarray.astype(np.uint8)

    img = Image.fromarray(image_ndarray, mode="RGB")
    img.save(filename, format="PNG")
  # ----------------------------------------------------------------------------------

  


_is_opencv_installed = importlib.util.find_spec("cv2") is not None
if _is_opencv_installed:
  import cv2

  class PNGFile_CV2(FileObject):
    # ----------------------------------------------------------------------------------
    def __init__(self, filename, parent_folder=None, error_template=None):
      super(PNGFile_CV2, self).__init__(filename, parent_folder, error_template, "png")
    # ----------------------------------------------------------------------------------
    def load(self, filename=None):
      filename = self._useFileName(filename)
      nImage = cv2.imread(filename)
      nImageRGB = cv2.cvtColor(nImage, cv2.COLOR_BGR2RGB)
      return nImageRGB
    # ----------------------------------------------------------------------------------
    def save(self, image_ndarray, filename=None):
      filename = self._useFileName(filename)
      cv2.imwrite(filename, image_ndarray)
    # ----------------------------------------------------------------------------------